/*-----------------------------------------------------------------------------
FILENAME: rfid.cpp
AUTHOR: S. Seegel, post@seegel-systeme.de
DATE: 22.7.2008
LICENSE: GPL v3
DESCRIPTION: Implementation of rfid class
-----------------------------------------------------------------------------*/

#include "rfid.h"
#include "global.h"
#include <avr/interrupt.h>
#include "tagmanager.h"

Rfid rfid;

#define F_RFID 125000UL	//RFID carrier frequency
#define TIMER0_PRESCALER 64 //Needed for manchester decoding

#define MANCHESTER_LONG (F_CPU / F_RFID * 64 / TIMER0_PRESCALER)
#define MANCHESTER_SHORT (F_CPU / F_RFID * 64 / TIMER0_PRESCALER / 2)
#define MANCHESTER_MIDDLE (MANCHESTER_LONG + MANCHESTER_SHORT) / 2

void Rfid::Init(void)
{
	#if ((HWVERSION == 1) || (HWVERSION == 2))
	SHDDDR |= 1<<SHDPINX;
	#endif
	
	TCCR0 = 1<<CS01 | 1<<CS00; //Timer/Counter0 prescaler 64
	MCUCR |= 1<<ISC00; //Any logical change on INT0 generates an interrupt request
	GICR |= 1<<INT0; //External interrupt request 0 enable (demod_out from RFID chip)
}

void Rfid::Execute(void)
{
	uint8_t ibit, obit, i, lp, p;
	
	if (rawtag_flag)
	{
		//calc line parities
		i=0;
		p=0;
		lp=0;
		for (ibit=9; ibit<59; ibit++)
		{
			lp ^= (rawtag[ibit / 8] >> (7 - (ibit % 8))) & 0x01;
			
			i++;
			if (i==5)
			{
				i=0;
				p |= lp;
			}
		}
			
		//calc line parities
		for (i=0; i<4; i++) //Check 4 columns
		{
			lp = 0;
			for (ibit=9+i; ibit<9+55+i; ibit+=5)
				lp ^= (rawtag[ibit / 8] >> (7 - (ibit % 8))) & 0x01;
			p |= lp;
		}
		
		//all parity bits ok ?
		if (!p)
		{
			tag[0] = 0; tag[1] = 0; tag[2] = 0; tag[3] = 0; tag[4] = 0;
			//Remove parity bits
			ibit = 9;
			for (obit=0; obit<40; obit++)
			{
				if ((rawtag[ibit / 8] >> (7 - (ibit % 8))) & 0x01)
					tag[obit / 8] |= 1 << (7 - (obit % 8));
				
				ibit++;
				if (((obit + 1) % 4) == 0)
					ibit++;
			}
			
			if (tag_flag == 0)
				tag_flag = 1;

			timer.SetTime(200);
		}
		
		rawtag_flag = 0;
		GICR |= 1<<INT0;
	}
	
	if (timer.IsFlagged())
	{
		tag_flag = 0;
	}
	
	if (tag_flag == 1)
	{
		tag_flag = 2;
		tagmanager.ProcessTag(tag);
	}
}

void Rfid::InjectByte(uint8_t b)
{
	if (bytecnt < 8)
	{
		rawtag[bytecnt++] = b;
		
		if (bytecnt == 8)
		{
			GICR &= ~(1<<INT0);
			rawtag_flag = 1;
		}
	}
}

void Rfid::TagStart(void)
{
	bytecnt = 0;
}



ISR(INT0_vect)
{
	static uint8_t old_cnt;
	static uint16_t sync;
	static uint8_t bitcount;
	static uint8_t insync = 0;
	
	if ((uint8_t)(TCNT0 - old_cnt) > MANCHESTER_MIDDLE)
	{
		old_cnt = TCNT0;
	
		sync <<= 1;
		bitcount++;
		
		if (bit_is_clear(PIND, PD2))
			sync |= 1;
		
		if (insync)
		{
			if ((bitcount % 8) == 0)
				rfid.InjectByte(sync);
			
			if (bitcount == 64)
				insync = 0;
		}
		else
			if ((sync & 0x3FF) == 0x1FF)
			{
				insync = 1;
				bitcount = 9;
				rfid.TagStart();
				rfid.InjectByte(sync);
			}
	}
}

